@@ -2,8 +2,10 @@ package sync
 
 import (
 	"context"
+	"encoding/json"
 	"fmt"
 	"sort"
+	"strconv"
 	"strings"
 	"sync"
 	"time"
@@ -21,7 +23,9 @@ import (
 	"k8s.io/client-go/rest"
 	"k8s.io/klog/v2/klogr"
 	cmdutil "k8s.io/kubectl/pkg/cmd/util"
+	"k8s.io/kubectl/pkg/util/openapi"
 
+	"github.com/argoproj/gitops-engine/pkg/diff"
 	"github.com/argoproj/gitops-engine/pkg/health"
 	"github.com/argoproj/gitops-engine/pkg/sync/common"
 	"github.com/argoproj/gitops-engine/pkg/sync/hook"
@@ -56,6 +60,13 @@ type SyncContext interface {
 // SyncOpt is a callback that update sync operation settings
 type SyncOpt func(ctx *syncContext)
 
+// WithPrunePropagationPolicy sets specified permission validator
+func WithPrunePropagationPolicy(policy *metav1.DeletionPropagation) SyncOpt {
+	return func(ctx *syncContext) {
+		ctx.prunePropagationPolicy = policy
+	}
+}
+
 // WithPermissionValidator sets specified permission validator
 func WithPermissionValidator(validator common.PermissionValidator) SyncOpt {
 	return func(ctx *syncContext) {
@@ -121,6 +132,25 @@ func WithManifestValidation(enabled bool) SyncOpt {
 	}
 }
 
+// WithPruneLast enables or disables pruneLast
+func WithPruneLast(enabled bool) SyncOpt {
+	return func(ctx *syncContext) {
+		ctx.pruneLast = enabled
+	}
+}
+
+// WithResourceModificationChecker sets resource modification result
+func WithResourceModificationChecker(enabled bool, diffResults *diff.DiffResultList) SyncOpt {
+	return func(ctx *syncContext) {
+		ctx.applyOutOfSyncOnly = enabled
+		if enabled {
+			ctx.modificationResult = groupDiffResults(diffResults)
+		} else {
+			ctx.modificationResult = nil
+		}
+	}
+}
+
 // WithNamespaceCreation will create non-exist namespace
 func WithNamespaceCreation(createNamespace bool, namespaceModifier func(*unstructured.Unstructured) bool) SyncOpt {
 	return func(ctx *syncContext) {
@@ -143,7 +173,13 @@ func WithSyncWaveHook(syncWaveHook common.SyncWaveHook) SyncOpt {
 	}
 }
 
-//  NewSyncContext creates new instance of a SyncContext
+func WithReplace(replace bool) SyncOpt {
+	return func(ctx *syncContext) {
+		ctx.replace = replace
+	}
+}
+
+// NewSyncContext creates new instance of a SyncContext
 func NewSyncContext(
 	revision string,
 	reconciliationResult ReconciliationResult,
@@ -151,19 +187,24 @@ func NewSyncContext(
 	rawConfig *rest.Config,
 	kubectl kubeutil.Kubectl,
 	namespace string,
+	openAPISchema openapi.Resources,
 	opts ...SyncOpt,
-) (SyncContext, error) {
+) (SyncContext, func(), error) {
 	dynamicIf, err := dynamic.NewForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	disco, err := discovery.NewDiscoveryClientForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	extensionsclientset, err := clientset.NewForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
+	}
+	resourceOps, cleanup, err := kubectl.ManageResources(rawConfig, openAPISchema)
+	if err != nil {
+		return nil, nil, err
 	}
 	ctx := &syncContext{
 		revision:            revision,
@@ -175,6 +216,7 @@ func NewSyncContext(
 		disco:               disco,
 		extensionsclientset: extensionsclientset,
 		kubectl:             kubectl,
+		resourceOps:         resourceOps,
 		namespace:           namespace,
 		log:                 klogr.New(),
 		validate:            true,
@@ -187,7 +229,7 @@ func NewSyncContext(
 	for _, opt := range opts {
 		opt(ctx)
 	}
-	return ctx, nil
+	return ctx, cleanup, nil
 }
 
 func groupResources(reconciliationResult ReconciliationResult) map[kubeutil.ResourceKey]reconciledResource {
@@ -209,6 +251,25 @@ func groupResources(reconciliationResult ReconciliationResult) map[kubeutil.Reso
 	return resources
 }
 
+// generates a map of resource and its modification result based on diffResultList
+func groupDiffResults(diffResultList *diff.DiffResultList) map[kubeutil.ResourceKey]bool {
+	modifiedResources := make(map[kube.ResourceKey]bool)
+	for _, res := range diffResultList.Diffs {
+		var obj unstructured.Unstructured
+		var err error
+		if string(res.NormalizedLive) != "null" {
+			err = json.Unmarshal(res.NormalizedLive, &obj)
+		} else {
+			err = json.Unmarshal(res.PredictedLive, &obj)
+		}
+		if err != nil {
+			continue
+		}
+		modifiedResources[kube.GetResourceKey(&obj)] = res.Modified
+	}
+	return modifiedResources
+}
+
 const (
 	crdReadinessTimeout = time.Duration(3) * time.Second
 )
@@ -249,14 +310,18 @@ type syncContext struct {
 	disco               discovery.DiscoveryInterface
 	extensionsclientset *clientset.Clientset
 	kubectl             kube.Kubectl
+	resourceOps         kube.ResourceOperations
 	namespace           string
 
-	dryRun          bool
-	force           bool
-	validate        bool
-	skipHooks       bool
-	resourcesFilter func(key kube.ResourceKey, target *unstructured.Unstructured, live *unstructured.Unstructured) bool
-	prune           bool
+	dryRun                 bool
+	force                  bool
+	validate               bool
+	skipHooks              bool
+	resourcesFilter        func(key kube.ResourceKey, target *unstructured.Unstructured, live *unstructured.Unstructured) bool
+	prune                  bool
+	replace                bool
+	pruneLast              bool
+	prunePropagationPolicy *metav1.DeletionPropagation
 
 	syncRes   map[string]common.ResourceSyncResult
 	startedAt time.Time
@@ -272,6 +337,10 @@ type syncContext struct {
 	namespaceModifier func(*unstructured.Unstructured) bool
 
 	syncWaveHook common.SyncWaveHook
+
+	applyOutOfSyncOnly bool
+	// stores whether the resource is modified or not
+	modificationResult map[kube.ResourceKey]bool
 }
 
 func (sc *syncContext) setRunningPhase(tasks []*syncTask, isPendingDeletion bool) {
@@ -313,8 +382,14 @@ func (sc *syncContext) Sync() {
 		// to perform the sync. we only wish to do this once per operation, performing additional dry-runs
 		// is harmless, but redundant. The indicator we use to detect if we have already performed
 		// the dry-run for this operation, is if the resource or hook list is empty.
-		sc.log.WithValues("tasks", tasks).Info("Tasks (dry-run)")
-		if sc.runTasks(tasks, true) == failed {
+
+		dryRunTasks := tasks
+		if sc.applyOutOfSyncOnly {
+			dryRunTasks = sc.filterOutOfSyncTasks(tasks)
+		}
+
+		sc.log.WithValues("tasks", dryRunTasks).Info("Tasks (dry-run)")
+		if sc.runTasks(dryRunTasks, true) == failed {
 			sc.setOperationPhase(common.OperationFailed, "one or more objects failed to apply (dry run)")
 			return
 		}
@@ -375,10 +450,12 @@ func (sc *syncContext) Sync() {
 	// syncFailTasks only run during failure, so separate them from regular tasks
 	syncFailTasks, tasks := tasks.Split(func(t *syncTask) bool { return t.phase == common.SyncPhaseSyncFail })
 
+	syncFailedTasks, _ := tasks.Split(func(t *syncTask) bool { return t.syncStatus == common.ResultCodeSyncFailed })
+
 	// if there are any completed but unsuccessful tasks, sync is a failure.
 	if tasks.Any(func(t *syncTask) bool { return t.completed() && !t.successful() }) {
 		sc.deleteHooks(hooksPendingDeletionFailed)
-		sc.setOperationFailed(syncFailTasks, "one or more synchronization tasks completed unsuccessfully")
+		sc.setOperationFailed(syncFailTasks, syncFailedTasks, "one or more synchronization tasks completed unsuccessfully")
 		return
 	}
 
@@ -386,6 +463,10 @@ func (sc *syncContext) Sync() {
 	// remove tasks that are completed, we can assume that there are no running tasks
 	tasks = tasks.Filter(func(t *syncTask) bool { return t.pending() })
 
+	if sc.applyOutOfSyncOnly {
+		tasks = sc.filterOutOfSyncTasks(tasks)
+	}
+
 	// If no sync tasks were generated (e.g., in case all application manifests have been removed),
 	// the sync operation is successful.
 	if len(tasks) == 0 {
@@ -425,8 +506,9 @@ func (sc *syncContext) Sync() {
 
 	switch runState {
 	case failed:
+		syncFailedTasks, _ := tasks.Split(func(t *syncTask) bool { return t.syncStatus == common.ResultCodeSyncFailed })
 		sc.deleteHooks(hooksPendingDeletionFailed)
-		sc.setOperationFailed(syncFailTasks, "one or more objects failed to apply")
+		sc.setOperationFailed(syncFailTasks, syncFailedTasks, "one or more objects failed to apply")
 	case successful:
 		if remainingTasks.Len() == 0 {
 			// delete all completed hooks which have appropriate delete policy
@@ -442,6 +524,21 @@ func (sc *syncContext) Sync() {
 	}
 }
 
+// filter out out-of-sync tasks
+func (sc *syncContext) filterOutOfSyncTasks(tasks syncTasks) syncTasks {
+	return tasks.Filter(func(t *syncTask) bool {
+		if t.isHook() {
+			return true
+		}
+
+		if modified, ok := sc.modificationResult[t.resourceKey()]; !modified && ok && t.targetObj != nil && t.liveObj != nil {
+			sc.log.WithValues("resource key", t.resourceKey()).V(1).Info("Skipping as resource was not modified")
+			return false
+		}
+		return true
+	})
+}
+
 func (sc *syncContext) deleteHooks(hooksPendingDeletion syncTasks) {
 	for _, task := range hooksPendingDeletion {
 		err := sc.deleteResource(task)
@@ -462,21 +559,33 @@ func (sc *syncContext) GetState() (common.OperationPhase, string, []common.Resou
 	return sc.phase, sc.message, resourceRes
 }
 
-func (sc *syncContext) setOperationFailed(syncFailTasks syncTasks, message string) {
+func (sc *syncContext) setOperationFailed(syncFailTasks, syncFailedTasks syncTasks, message string) {
+	errorMessageFactory := func(tasks []*syncTask, message string) string {
+		messages := syncFailedTasks.Map(func(task *syncTask) string {
+			return task.message
+		})
+		if len(messages) > 0 {
+			return fmt.Sprintf("%s, reason: %s", message, strings.Join(messages, ","))
+		}
+		return message
+	}
+
+	errorMessage := errorMessageFactory(syncFailedTasks, message)
+
 	if len(syncFailTasks) > 0 {
 		// if all the failure hooks are completed, don't run them again, and mark the sync as failed
 		if syncFailTasks.All(func(task *syncTask) bool { return task.completed() }) {
-			sc.setOperationPhase(common.OperationFailed, message)
+			sc.setOperationPhase(common.OperationFailed, errorMessage)
 			return
 		}
 		// otherwise, we need to start the failure hooks, and then return without setting
 		// the phase, so we make sure we have at least one more sync
 		sc.log.WithValues("syncFailTasks", syncFailTasks).V(1).Info("Running sync fail tasks")
 		if sc.runTasks(syncFailTasks, false) == failed {
-			sc.setOperationPhase(common.OperationFailed, message)
+			sc.setOperationPhase(common.OperationFailed, errorMessage)
 		}
 	} else {
-		sc.setOperationPhase(common.OperationFailed, message)
+		sc.setOperationPhase(common.OperationFailed, errorMessage)
 	}
 }
 
@@ -597,6 +706,29 @@ func (sc *syncContext) getSyncTasks() (_ syncTasks, successful bool) {
 		}
 	}
 
+	// for pruneLast tasks, modify the wave to sync phase last wave of non prune task +1
+	syncPhaseLastWave := 0
+	for _, task := range tasks {
+		if task.phase == common.SyncPhaseSync {
+			if task.wave() > syncPhaseLastWave && !task.isPrune() {
+				syncPhaseLastWave = task.wave()
+			}
+		}
+	}
+	syncPhaseLastWave = syncPhaseLastWave + 1
+
+	for _, task := range tasks {
+		if task.isPrune() &&
+			(sc.pruneLast || resourceutil.HasAnnotationOption(task.liveObj, common.AnnotationSyncOptions, common.SyncOptionPruneLast)) {
+			annotations := task.liveObj.GetAnnotations()
+			if annotations == nil {
+				annotations = make(map[string]string)
+			}
+			annotations[common.AnnotationSyncWave] = strconv.Itoa(syncPhaseLastWave)
+			task.liveObj.SetAnnotations(annotations)
+		}
+	}
+
 	tasks.Sort()
 
 	// finally enrich tasks with the result
@@ -666,7 +798,6 @@ func isNamespaceWithName(res *unstructured.Unstructured, ns string) bool {
 
 func isNamespaceKind(res *unstructured.Unstructured) bool {
 	return res != nil &&
-		res.GetObjectKind() != nil &&
 		res.GetObjectKind().GroupVersionKind().Group == "" &&
 		res.GetKind() == kube.NamespaceKind
 }
@@ -716,17 +847,41 @@ func (sc *syncContext) ensureCRDReady(name string) {
 	})
 }
 
-func (sc *syncContext) applyObject(targetObj *unstructured.Unstructured, dryRun bool, force bool, validate bool) (common.ResultCode, string) {
+func (sc *syncContext) applyObject(t *syncTask, dryRun bool, force bool, validate bool) (common.ResultCode, string) {
 	dryRunStrategy := cmdutil.DryRunNone
 	if dryRun {
 		dryRunStrategy = cmdutil.DryRunClient
 	}
-	message, err := sc.kubectl.ApplyResource(context.TODO(), sc.rawConfig, targetObj, targetObj.GetNamespace(), dryRunStrategy, force, validate)
+
+	var err error
+	var message string
+	shouldReplace := sc.replace || resourceutil.HasAnnotationOption(t.targetObj, common.AnnotationSyncOptions, common.SyncOptionReplace)
+	if shouldReplace {
+		if t.liveObj != nil {
+			// Avoid using `kubectl replace` for CRDs since 'replace' might recreate resource and so delete all CRD instances
+			if kube.IsCRD(t.targetObj) {
+				update := t.targetObj.DeepCopy()
+				update.SetResourceVersion(t.liveObj.GetResourceVersion())
+				_, err = sc.resourceOps.UpdateResource(context.TODO(), update, dryRunStrategy)
+				if err == nil {
+					message = fmt.Sprintf("%s/%s updated", t.targetObj.GetKind(), t.targetObj.GetName())
+				} else {
+					message = fmt.Sprintf("error when updating: %v", err.Error())
+				}
+			} else {
+				message, err = sc.resourceOps.ReplaceResource(context.TODO(), t.targetObj, dryRunStrategy, force)
+			}
+		} else {
+			message, err = sc.resourceOps.CreateResource(context.TODO(), t.targetObj, dryRunStrategy, validate)
+		}
+	} else {
+		message, err = sc.resourceOps.ApplyResource(context.TODO(), t.targetObj, dryRunStrategy, force, validate)
+	}
 	if err != nil {
 		return common.ResultCodeSyncFailed, err.Error()
 	}
-	if kube.IsCRD(targetObj) && !dryRun {
-		sc.ensureCRDReady(targetObj.GetName())
+	if kube.IsCRD(t.targetObj) && !dryRun {
+		sc.ensureCRDReady(t.targetObj.GetName())
 	}
 	return common.ResultCodeSynced, message
 }
@@ -744,7 +899,7 @@ func (sc *syncContext) pruneObject(liveObj *unstructured.Unstructured, prune, dr
 			// Skip deletion if object is already marked for deletion, so we don't cause a resource update hotloop
 			deletionTimestamp := liveObj.GetDeletionTimestamp()
 			if deletionTimestamp == nil || deletionTimestamp.IsZero() {
-				err := sc.kubectl.DeleteResource(context.TODO(), sc.config, liveObj.GroupVersionKind(), liveObj.GetName(), liveObj.GetNamespace(), false)
+				err := sc.kubectl.DeleteResource(context.TODO(), sc.config, liveObj.GroupVersionKind(), liveObj.GetName(), liveObj.GetNamespace(), sc.getDeleteOptions())
 				if err != nil {
 					return common.ResultCodeSyncFailed, err.Error()
 				}
@@ -754,6 +909,15 @@ func (sc *syncContext) pruneObject(liveObj *unstructured.Unstructured, prune, dr
 	}
 }
 
+func (sc *syncContext) getDeleteOptions() metav1.DeleteOptions {
+	propagationPolicy := metav1.DeletePropagationForeground
+	if sc.prunePropagationPolicy != nil {
+		propagationPolicy = *sc.prunePropagationPolicy
+	}
+	deleteOption := metav1.DeleteOptions{PropagationPolicy: &propagationPolicy}
+	return deleteOption
+}
+
 func (sc *syncContext) targetObjs() []*unstructured.Unstructured {
 	objs := sc.hooks
 	for _, r := range sc.resources {
@@ -829,8 +993,7 @@ func (sc *syncContext) deleteResource(task *syncTask) error {
 	if err != nil {
 		return err
 	}
-	propagationPolicy := metav1.DeletePropagationForeground
-	return resIf.Delete(context.TODO(), task.name(), metav1.DeleteOptions{PropagationPolicy: &propagationPolicy})
+	return resIf.Delete(context.TODO(), task.name(), sc.getDeleteOptions())
 }
 
 func (sc *syncContext) getResourceIf(task *syncTask) (dynamic.ResourceInterface, error) {
@@ -962,13 +1125,19 @@ func (sc *syncContext) processCreateTasks(state runState, tasks syncTasks, dryRu
 			logCtx := sc.log.WithValues("dryRun", dryRun, "task", t)
 			logCtx.V(1).Info("Applying")
 			validate := sc.validate && !resourceutil.HasAnnotationOption(t.targetObj, common.AnnotationSyncOptions, common.SyncOptionsDisableValidation)
-			result, message := sc.applyObject(t.targetObj, dryRun, sc.force, validate)
+			result, message := sc.applyObject(t, dryRun, sc.force, validate)
 			if result == common.ResultCodeSyncFailed {
 				logCtx.WithValues("message", message).Info("Apply failed")
 				state = failed
 			}
 			if !dryRun || sc.dryRun || result == common.ResultCodeSyncFailed {
-				sc.setResourceResult(t, result, operationPhases[result], message)
+				phase := operationPhases[result]
+				// no resources are created in dry-run, so running phase means validation was
+				// successful and sync operation succeeded
+				if sc.dryRun && phase == common.OperationRunning {
+					phase = common.OperationSucceeded
+				}
+				sc.setResourceResult(t, result, phase, message)
 			}
 			return state
 		})
