@@ -15,15 +15,16 @@
 package casbin
 
 import (
+	"fmt"
 	"testing"
 
-	"github.com/casbin/casbin/v2/persist/file-adapter"
+	"github.com/casbin/casbin/v2/log"
+	fileadapter "github.com/casbin/casbin/v2/persist/file-adapter"
 	"github.com/casbin/casbin/v2/rbac"
-	"github.com/casbin/casbin/v2/rbac/default-role-manager"
 	"github.com/casbin/casbin/v2/util"
 )
 
-func testEnforce(t *testing.T, e *Enforcer, sub string, obj interface{}, act string, res bool) {
+func testEnforce(t *testing.T, e *Enforcer, sub interface{}, obj interface{}, act string, res bool) {
 	t.Helper()
 	if myRes, _ := e.Enforce(sub, obj, act); myRes != res {
 		t.Errorf("%s, %v, %s: %t, supposed to be %t", sub, obj, act, myRes, res)
@@ -57,6 +58,19 @@ func TestBasicModel(t *testing.T) {
 	testEnforce(t, e, "bob", "data2", "write", true)
 }
 
+func TestBasicModelWithoutSpaces(t *testing.T) {
+	e, _ := NewEnforcer("examples/basic_model_without_spaces.conf", "examples/basic_policy.csv")
+
+	testEnforce(t, e, "alice", "data1", "read", true)
+	testEnforce(t, e, "alice", "data1", "write", false)
+	testEnforce(t, e, "alice", "data2", "read", false)
+	testEnforce(t, e, "alice", "data2", "write", false)
+	testEnforce(t, e, "bob", "data1", "read", false)
+	testEnforce(t, e, "bob", "data1", "write", false)
+	testEnforce(t, e, "bob", "data2", "read", false)
+	testEnforce(t, e, "bob", "data2", "write", true)
+}
+
 func TestBasicModelNoPolicy(t *testing.T) {
 	e, _ := NewEnforcer("examples/basic_model.conf")
 
@@ -164,13 +178,13 @@ func TestRBACModelWithDomains(t *testing.T) {
 func TestRBACModelWithDomainsAtRuntime(t *testing.T) {
 	e, _ := NewEnforcer("examples/rbac_with_domains_model.conf")
 
-	e.AddPolicy("admin", "domain1", "data1", "read")
-	e.AddPolicy("admin", "domain1", "data1", "write")
-	e.AddPolicy("admin", "domain2", "data2", "read")
-	e.AddPolicy("admin", "domain2", "data2", "write")
+	_, _ = e.AddPolicy("admin", "domain1", "data1", "read")
+	_, _ = e.AddPolicy("admin", "domain1", "data1", "write")
+	_, _ = e.AddPolicy("admin", "domain2", "data2", "read")
+	_, _ = e.AddPolicy("admin", "domain2", "data2", "write")
 
-	e.AddGroupingPolicy("alice", "admin", "domain1")
-	e.AddGroupingPolicy("bob", "admin", "domain2")
+	_, _ = e.AddGroupingPolicy("alice", "admin", "domain1")
+	_, _ = e.AddGroupingPolicy("bob", "admin", "domain2")
 
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", true)
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "write", true)
@@ -182,7 +196,7 @@ func TestRBACModelWithDomainsAtRuntime(t *testing.T) {
 	testDomainEnforce(t, e, "bob", "domain2", "data2", "write", true)
 
 	// Remove all policy rules related to domain1 and data1.
-	e.RemoveFilteredPolicy(1, "domain1", "data1")
+	_, _ = e.RemoveFilteredPolicy(1, "domain1", "data1")
 
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", false)
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "write", false)
@@ -194,7 +208,7 @@ func TestRBACModelWithDomainsAtRuntime(t *testing.T) {
 	testDomainEnforce(t, e, "bob", "domain2", "data2", "write", true)
 
 	// Remove the specified policy rule.
-	e.RemovePolicy("admin", "domain2", "data2", "read")
+	_, _ = e.RemovePolicy("admin", "domain2", "data2", "read")
 
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", false)
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "write", false)
@@ -210,17 +224,17 @@ func TestRBACModelWithDomainsAtRuntimeMockAdapter(t *testing.T) {
 	adapter := fileadapter.NewAdapterMock("examples/rbac_with_domains_policy.csv")
 	e, _ := NewEnforcer("examples/rbac_with_domains_model.conf", adapter)
 
-	e.AddPolicy("admin", "domain3", "data1", "read")
-	e.AddGroupingPolicy("alice", "admin", "domain3")
+	_, _ = e.AddPolicy("admin", "domain3", "data1", "read")
+	_, _ = e.AddGroupingPolicy("alice", "admin", "domain3")
 
 	testDomainEnforce(t, e, "alice", "domain3", "data1", "read", true)
 
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", true)
-	e.RemoveFilteredPolicy(1, "domain1", "data1")
+	_, _ = e.RemoveFilteredPolicy(1, "domain1", "data1")
 	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", false)
 
 	testDomainEnforce(t, e, "bob", "domain2", "data2", "read", true)
-	e.RemovePolicy("admin", "domain2", "data2", "read")
+	_, _ = e.RemovePolicy("admin", "domain2", "data2", "read")
 	testDomainEnforce(t, e, "bob", "domain2", "data2", "read", false)
 }
 
@@ -249,7 +263,7 @@ func TestRBACModelWithCustomData(t *testing.T) {
 	// You can add custom data to a grouping policy, Casbin will ignore it. It is only meaningful to the caller.
 	// This feature can be used to store information like whether "bob" is an end user (so no subject will inherit "bob")
 	// For Casbin, it is equivalent to: e.AddGroupingPolicy("bob", "data2_admin")
-	e.AddGroupingPolicy("bob", "data2_admin", "custom_data")
+	_, _ = e.AddGroupingPolicy("bob", "data2_admin", "custom_data")
 
 	testEnforce(t, e, "alice", "data1", "read", true)
 	testEnforce(t, e, "alice", "data1", "write", false)
@@ -263,7 +277,7 @@ func TestRBACModelWithCustomData(t *testing.T) {
 	// You should also take the custom data as a parameter when deleting a grouping policy.
 	// e.RemoveGroupingPolicy("bob", "data2_admin") won't work.
 	// Or you can remove it by using RemoveFilteredGroupingPolicy().
-	e.RemoveGroupingPolicy("bob", "data2_admin", "custom_data")
+	_, _ = e.RemoveGroupingPolicy("bob", "data2_admin", "custom_data")
 
 	testEnforce(t, e, "alice", "data1", "read", true)
 	testEnforce(t, e, "alice", "data1", "write", false)
@@ -283,7 +297,7 @@ func TestRBACModelWithPattern(t *testing.T) {
 	// You can see in policy that: "g2, /book/:id, book_group", so in "g2()" function in the matcher, instead
 	// of checking whether "/book/:id" equals the obj: "/book/1", it checks whether the pattern matches.
 	// You can see it as normal RBAC: "/book/:id" == "/book/1" becomes KeyMatch2("/book/:id", "/book/1")
-	e.rm.(*defaultrolemanager.RoleManager).AddMatchingFunc("KeyMatch2", util.KeyMatch2)
+	e.AddNamedMatchingFunc("g2", "KeyMatch2", util.KeyMatch2)
 	testEnforce(t, e, "alice", "/book/1", "GET", true)
 	testEnforce(t, e, "alice", "/book/2", "GET", true)
 	testEnforce(t, e, "alice", "/pen/1", "GET", true)
@@ -295,7 +309,7 @@ func TestRBACModelWithPattern(t *testing.T) {
 
 	// AddMatchingFunc() is actually setting a function because only one function is allowed,
 	// so when we set "KeyMatch3", we are actually replacing "KeyMatch2" with "KeyMatch3".
-	e.rm.(*defaultrolemanager.RoleManager).AddMatchingFunc("KeyMatch3", util.KeyMatch3)
+	e.AddNamedMatchingFunc("g2", "KeyMatch2", util.KeyMatch3)
 	testEnforce(t, e, "alice", "/book2/1", "GET", true)
 	testEnforce(t, e, "alice", "/book2/2", "GET", true)
 	testEnforce(t, e, "alice", "/pen2/1", "GET", true)
@@ -315,6 +329,9 @@ func (rm *testCustomRoleManager) Clear() error { return nil }
 func (rm *testCustomRoleManager) AddLink(name1 string, name2 string, domain ...string) error {
 	return nil
 }
+func (rm *testCustomRoleManager) BuildRelationship(name1 string, name2 string, domain ...string) error {
+	return nil
+}
 func (rm *testCustomRoleManager) DeleteLink(name1 string, name2 string, domain ...string) error {
 	return nil
 }
@@ -334,12 +351,16 @@ func (rm *testCustomRoleManager) GetRoles(name string, domain ...string) ([]stri
 func (rm *testCustomRoleManager) GetUsers(name string, domain ...string) ([]string, error) {
 	return []string{}, nil
 }
-func (rm *testCustomRoleManager) PrintRoles() error { return nil }
+func (rm *testCustomRoleManager) GetDomains(name string) ([]string, error) {
+	return []string{}, nil
+}
+func (rm *testCustomRoleManager) PrintRoles() error           { return nil }
+func (rm *testCustomRoleManager) SetLogger(logger log.Logger) {}
 
 func TestRBACModelWithCustomRoleManager(t *testing.T) {
 	e, _ := NewEnforcer("examples/rbac_model.conf", "examples/rbac_policy.csv")
 	e.SetRoleManager(NewRoleManager())
-	e.LoadModel()
+	_ = e.LoadModel()
 	_ = e.LoadPolicy()
 
 	testEnforce(t, e, "alice", "data1", "read", true)
@@ -465,6 +486,25 @@ func TestIPMatchModel(t *testing.T) {
 	testEnforce(t, e, "192.168.0.1", "data2", "write", false)
 }
 
+func TestGlobMatchModel(t *testing.T) {
+	e, _ := NewEnforcer("examples/glob_model.conf", "examples/glob_policy.csv")
+	testEnforce(t, e, "u1", "/foo/", "read", true)
+	testEnforce(t, e, "u1", "/foo", "read", false)
+	testEnforce(t, e, "u1", "/foo/subprefix", "read", true)
+	testEnforce(t, e, "u1", "foo", "read", false)
+
+	testEnforce(t, e, "u2", "/foosubprefix", "read", true)
+	testEnforce(t, e, "u2", "/foo/subprefix", "read", false)
+	testEnforce(t, e, "u2", "foo", "read", false)
+
+	testEnforce(t, e, "u3", "/prefix/foo/subprefix", "read", true)
+	testEnforce(t, e, "u3", "/prefix/foo/", "read", true)
+	testEnforce(t, e, "u3", "/prefix/foo", "read", false)
+
+	testEnforce(t, e, "u4", "/foo", "read", false)
+	testEnforce(t, e, "u4", "foo", "read", true)
+}
+
 func TestPriorityModel(t *testing.T) {
 	e, _ := NewEnforcer("examples/priority_model.conf", "examples/priority_policy.csv")
 
@@ -496,3 +536,92 @@ func TestRBACModelInMultiLines(t *testing.T) {
 	testEnforce(t, e, "bob", "data2", "read", false)
 	testEnforce(t, e, "bob", "data2", "write", true)
 }
+
+type testSub struct {
+	Name string
+	Age  int
+}
+
+func newTestSubject(name string, age int) testSub {
+	s := testSub{}
+	s.Name = name
+	s.Age = age
+	return s
+}
+
+func TestABACNotUsingPolicy(t *testing.T) {
+	e, _ := NewEnforcer("examples/abac_not_using_policy_model.conf", "examples/abac_rule_effect_policy.csv")
+	data1 := newTestResource("data1", "alice")
+	data2 := newTestResource("data2", "bob")
+
+	testEnforce(t, e, "alice", data1, "read", true)
+	testEnforce(t, e, "alice", data1, "write", true)
+	testEnforce(t, e, "alice", data2, "read", false)
+	testEnforce(t, e, "alice", data2, "write", false)
+}
+
+func TestABACPolicy(t *testing.T) {
+	e, _ := NewEnforcer("examples/abac_rule_model.conf", "examples/abac_rule_policy.csv")
+	m := e.GetModel()
+	for sec, ast := range m {
+		fmt.Println(sec)
+		for ptype, p := range ast {
+			fmt.Println(ptype, p)
+		}
+	}
+	sub1 := newTestSubject("alice", 16)
+	sub2 := newTestSubject("alice", 20)
+	sub3 := newTestSubject("alice", 65)
+
+	testEnforce(t, e, sub1, "/data1", "read", false)
+	testEnforce(t, e, sub1, "/data2", "read", false)
+	testEnforce(t, e, sub1, "/data1", "write", false)
+	testEnforce(t, e, sub1, "/data2", "write", true)
+	testEnforce(t, e, sub2, "/data1", "read", true)
+	testEnforce(t, e, sub2, "/data2", "read", false)
+	testEnforce(t, e, sub2, "/data1", "write", false)
+	testEnforce(t, e, sub2, "/data2", "write", true)
+	testEnforce(t, e, sub3, "/data1", "read", true)
+	testEnforce(t, e, sub3, "/data2", "read", false)
+	testEnforce(t, e, sub3, "/data1", "write", false)
+	testEnforce(t, e, sub3, "/data2", "write", false)
+}
+
+func TestCommentModel(t *testing.T) {
+	e, _ := NewEnforcer("examples/comment_model.conf", "examples/basic_policy.csv")
+	testEnforce(t, e, "alice", "data1", "read", true)
+	testEnforce(t, e, "alice", "data1", "write", false)
+	testEnforce(t, e, "alice", "data2", "read", false)
+	testEnforce(t, e, "alice", "data2", "write", false)
+	testEnforce(t, e, "bob", "data1", "read", false)
+	testEnforce(t, e, "bob", "data1", "write", false)
+	testEnforce(t, e, "bob", "data2", "read", false)
+	testEnforce(t, e, "bob", "data2", "write", true)
+}
+
+func TestDomainMatchModel(t *testing.T) {
+	e, _ := NewEnforcer("examples/rbac_with_domain_pattern_model.conf", "examples/rbac_with_domain_pattern_policy.csv")
+	e.AddNamedDomainMatchingFunc("g", "keyMatch2", util.KeyMatch2)
+
+	testDomainEnforce(t, e, "alice", "domain1", "data1", "read", true)
+	testDomainEnforce(t, e, "alice", "domain1", "data1", "write", true)
+	testDomainEnforce(t, e, "alice", "domain1", "data2", "read", false)
+	testDomainEnforce(t, e, "alice", "domain1", "data2", "write", false)
+	testDomainEnforce(t, e, "alice", "domain2", "data2", "read", true)
+	testDomainEnforce(t, e, "alice", "domain2", "data2", "write", true)
+	testDomainEnforce(t, e, "bob", "domain2", "data1", "read", false)
+	testDomainEnforce(t, e, "bob", "domain2", "data1", "write", false)
+	testDomainEnforce(t, e, "bob", "domain2", "data2", "read", true)
+	testDomainEnforce(t, e, "bob", "domain2", "data2", "write", true)
+}
+
+func TestAllMatchModel(t *testing.T) {
+	e, _ := NewEnforcer("examples/rbac_with_all_pattern_model.conf", "examples/rbac_with_all_pattern_policy.csv")
+	e.AddNamedMatchingFunc("g", "keyMatch2", util.KeyMatch2)
+	e.AddNamedDomainMatchingFunc("g", "keyMatch2", util.KeyMatch2)
+
+	testDomainEnforce(t, e, "alice", "domain1", "/book/1", "read", true)
+	testDomainEnforce(t, e, "alice", "domain1", "/book/1", "write", false)
+	testDomainEnforce(t, e, "alice", "domain2", "/book/1", "read", false)
+	testDomainEnforce(t, e, "alice", "domain2", "/book/1", "write", true)
+}
